
/*
 * dialog.c --
 *
 * This file implements the gnome dialog and messageBox widget
 *
 * Copyright (c) 2001 - 2002 P.G. Baum
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
   History:
   2002-05: moved from gnome dialog to gtk 2.0 dialog
   2001-05: Begin of developement
 */

#include "gnocl.h"
#include <string.h>
#include <assert.h>

typedef struct
{
   Gnocl_StringBool child;    /* either child or text */
   Gnocl_StringBool text;
   Gnocl_StringBool title;
   Gnocl_StringBool parent;
   Gnocl_IntBool    defaultBut;
   Gnocl_ObjBool    type;
   Gnocl_BoolBool   modal;
} DialogParams;

static const GnoclWidgetOptions dialogOptions[] =
{
   { "-child", GNOCL_STRING, GNOCL_OFFSET( DialogParams, child ) },
   { "-title", GNOCL_STRING, GNOCL_OFFSET( DialogParams, title ) },
   { "-text", GNOCL_STRING, GNOCL_OFFSET( DialogParams, text ) },
   { "-parent", GNOCL_STRING, GNOCL_OFFSET( DialogParams, parent ) },
   { "-default", GNOCL_INT, GNOCL_OFFSET( DialogParams, defaultBut ) },
   { "-type", GNOCL_OBJ, GNOCL_OFFSET( DialogParams, type ) },
   { "-modal", GNOCL_BOOL, GNOCL_OFFSET( DialogParams, modal ) },
   { NULL, 0, 0 }
};

int gnoclDialogCmd( ClientData data, Tcl_Interp *interp,
      int objc, Tcl_Obj * const objv[] )
{
   char *typeNames[] = { "none", "info", "warning", "question", "error", NULL };
   const int types[] = {  -1, GTK_MESSAGE_INFO, GTK_MESSAGE_WARNING, 
         GTK_MESSAGE_QUESTION, GTK_MESSAGE_ERROR };

   char *butNames[] = { GNOCL_STOCK_PREFIX "None", GNOCL_STOCK_PREFIX "Ok", 
         GNOCL_STOCK_PREFIX "Close", GNOCL_STOCK_PREFIX "Cancel", 
         GNOCL_STOCK_PREFIX "YesNo", GNOCL_STOCK_PREFIX "OkCancel", NULL };
   const GtkButtonsType butTypes[] = { GTK_BUTTONS_NONE, GTK_BUTTONS_OK,
         GTK_BUTTONS_CLOSE, GTK_BUTTONS_CANCEL, GTK_BUTTONS_YES_NO,
         GTK_BUTTONS_OK_CANCEL };

   DialogParams pp;
   GtkDialog    *dialog;
   int          n, noButtons;
   GtkWidget    *child = NULL;
   GtkWidget    *parent = NULL;
   int          ret;
   int          buttonHandled = 0;
   int typeIdx = 0;

   if( objc < 2 )
   {
      Tcl_WrongNumArgs( interp, 1, objv, "buttons ?option val ...?" );
      return TCL_ERROR;
   }
   memset( &pp, 0, sizeof( pp ) );
   if( gnoclParseAllWidgetOpts( interp, objc, objv, 1,
         dialogOptions, (char *)&pp ) != TCL_OK )
      return TCL_ERROR;

   if( Tcl_ListObjLength( interp, objv[1], &noButtons ) != TCL_OK 
         || noButtons < 1 )
   {
      Tcl_SetResult( interp, "button-list must be proper list", 
            TCL_STATIC );
      return TCL_ERROR;
   }
   if( pp.text.changed && pp.child.changed )
   {
      Tcl_SetResult( interp, "either \"text\" or \"child\" may be specified", 
            TCL_STATIC );
      return TCL_ERROR;
   }
   
   if( pp.child.changed )
   {
      child = gnoclGetWidgetFromName( pp.child.val, interp );
      if( child == NULL )
         return TCL_ERROR;
   }

   if( pp.parent.changed )
   {
      GtkWidget *par = gnoclGetWidgetFromName( pp.parent.val, interp );
      if( par == NULL )
         return TCL_ERROR;

      parent = GTK_WIDGET( par );
      /* TODO: test this case */
   }
   if( pp.type.changed  )
      if( Tcl_GetIndexFromObj( interp, pp.type.val, typeNames, "type",
            TCL_EXACT, &typeIdx ) != TCL_OK )
         return TCL_ERROR;

   if( typeIdx == 0 )
   {
      dialog = GTK_DIALOG( gtk_dialog_new( ) );
      if( pp.text.changed )
         child = gtk_label_new( pp.text.val );
   }
   else
   {
      GtkButtonsType butType = GTK_BUTTONS_NONE;
      char *txt = "";

      if( pp.text.changed )
         txt = pp.text.val;

      if( noButtons == 1  && 
            (gnoclGetStringType( objv[1] ) & GNOCL_STR_STOCK )  )
      {
         int idx;
         if( Tcl_GetIndexFromObj( NULL, objv[1], butNames, "",
               TCL_EXACT, &idx ) == TCL_OK )
         {
            butType = butTypes[idx];
            buttonHandled = 1;
         }
      }
      dialog = GTK_DIALOG( gtk_message_dialog_new( NULL,
            0, types[typeIdx], butType, "%s", txt ) );
   }

   if( pp.title.changed )
      gtk_window_set_title( GTK_WINDOW( dialog ), pp.title.val );

   if( buttonHandled == 0 )
   {
      for( n = 0; n < noButtons; ++n )
      {
         Tcl_Obj *pobj;
         char *but = NULL;
         GnoclStringType st;

         if( Tcl_ListObjIndex( interp, objv[1], n, &pobj ) != TCL_OK )
            return TCL_ERROR;
         st = gnoclGetStringType( pobj );
         if( st & GNOCL_STR_STOCK )
         {
            but = gnoclGetStockName( pobj, interp, NULL );
            if( but == NULL )
               return TCL_ERROR;
         }
         else
            but = gnoclGetString( pobj );
         gtk_dialog_add_button( dialog, but, n );
         if( st & GNOCL_STR_STOCK )
            g_free( but );
      }
   }


   if( child )
   {
      gtk_box_pack_start( GTK_BOX( dialog->vbox ), child, TRUE, TRUE, 0 );
      gtk_widget_show_all( GTK_WIDGET( dialog->vbox ) );
   }

   if( parent )
      gtk_widget_set_parent( GTK_WIDGET( dialog ), parent );

   if( pp.defaultBut.changed && pp.defaultBut.val >= 0 )
      gtk_dialog_set_default_response( dialog,  pp.defaultBut.val );

   if( pp.modal.changed )
      gtk_window_set_modal( GTK_WINDOW( dialog ), pp.modal.val );

   ret = gtk_dialog_run( dialog );
   gtk_widget_destroy( GTK_WIDGET( dialog ) );

   /* TODO: translate ret to "ok", "cancel", ... */
   Tcl_SetObjResult( interp, Tcl_NewIntObj( ret ) );

   return TCL_OK;
}

